#include "option.h"

namespace OptionLibrary
{
	/*
		c_copy : stringchar*ؤѴ

		const string& s : оʸ
		char* c : Ѵ̤ϿcharؤΥݥ
		size_t size : cΥ
	*/
	void c_copy( const string& s, char* c, size_t size )
	{
		if ( s.length() == 0 || size == 0 ) return;
		size_t i = ( s.length() + 1 > size ) ? size - 1 : s.length();
		s.copy( c, i );
		c[i] = 0;
	}

	/*
		ʸ󤫤intؤѴ

		const string& s : оʸ
		int& value : ͤǼѿ
	*/
	bool toInt( const string& s, int& value )
	{
		char* c = new char[s.length() + 1];
		char* e;

		c_copy( s, c, s.length() + 1 );
		value = strtol( c, &e, 10 );

		bool b = *c != '\0' && *e == '\0';
		if ( ! b ) value = 0;

		delete[] c;

		return ( b );
	}

	/*
		ʸ󤫤unsigned intؤѴ

		const string& s : оʸ
		unsigned int& value : ͤǼѿ
	*/
	bool toUInt( const string& s, unsigned int& value )
	{
		char* c = new char[s.length() + 1];
		char* e;

		c_copy( s, c, s.length() + 1 );
		value = strtoul( c, &e, 10 );

		bool b = *c != '\0' && *e == '\0';
		if ( ! b ) value = 0;

		delete[] c;

		return ( b );
	}

	/*
		ʸ󤫤charؤѴ

		const string& s : оʸ
		char& value : ͤǼѿ
	*/
	bool toChar( const string& s, char& value )
	{
		int i;
		if ( ! toInt( s, i ) ) return( false );

		if ( i < CHAR_MIN || i > CHAR_MAX ) return( false );
		value = (char)i;

		return( true );
	}

	/*
		ʸ󤫤unsigned charؤѴ

		const string& s : оʸ
		unsigned char& value : ͤǼѿ
	*/
	bool toUChar( const string& s, unsigned char& value )
	{
		unsigned int i;
		if ( ! toUInt( s, i ) ) return( false );

		if ( i > UCHAR_MAX ) return( false );
		value = (unsigned char)i;

		return( true );
	}

	/*
		ʸ󤫤doubleؤѴ

		const string& s : оʸ
		double& value : ͤǼѿ
	*/
	bool toDouble( const string& s, double& value )
	{
		char* c = new char[s.length() + 1];
		char* e;

		c_copy( s, c, s.length() + 1 );
		value = strtod( c, &e );

		bool b = *c != '\0' && *e == '\0';
		if ( ! b ) value = 0;

		delete[] c;

		return ( b );
	}

	/*
		Args󥹥ȥ饯

		DefaultOptionValue optionArray[] : ץȥǥեͤ
		size_t optionSize : optionArrayΥ
	*/
	Args::Args( DefaultOptionValue optionArray[], size_t optionSize )
	{
		for ( unsigned int i = 0 ; i < optionSize ; i++ ) {
			DefaultOptionValue& o = optionArray[i];
			if ( o.ident.length() == 0 )
				continue;
			string keyList = o.key;
			string::size_type st;
			do {
				st = keyList.find( ',' );
				string key = keyList.substr( 0, st );
				if ( st != string::npos ) keyList = keyList.substr( st + 1 );
				if ( optKey.find( key ) == optKey.end() )
					optKey[key] = o.ident;
			} while ( st != string::npos );
			if ( option.find( o.ident ) == option.end() ) {
				OptionValue newVal( o );
				option[o.ident] = newVal;
			}
		}
	}

	/*
		Args::set βϤϿ

		int argc : θĿ
		char* argv[] : Υꥹ
	*/
	bool Args::set( int argc, char* argv[] )
	{
		string curIdent = "";
		bool ans = true;

		for ( int i = 1 ; i < argc ; i++ ) {
			if ( argv[i][0] == '-' ) {
				string key = &argv[i][1];
				string::size_type st = key.find( '=' );
				if ( st != string::npos ) key = key.substr( 0, st );
				map<string,string>::const_iterator kit = optKey.find( key );
				if ( kit == optKey.end() ) {	// ץ󥭡Ĥʤä
					fprintf( stderr, "Option key %s is not found. It's ignored.\n", &argv[i][1] );
					ans = false;
				} else {
					MapOpt::iterator it = option.find( kit->second );
					if ( curIdent != "" ) {	// ͤɬפʥФͤꤵƤʤä
						fprintf( stderr, "The value for option key %s is not set. It's ignored.\n", curIdent.c_str() );
						curIdent = "";
						ans = false;
					}
					if ( ! ( it->second ).hasValue ) {	// ͤפʥץ
						( it->second ).set();
					} else {	// ͤɬפʥץ
						if ( st != string::npos ) {
							OptionValue& o = option[kit->second];
							string s = argv[i];
							o.set( s.substr( s.find( '=' ) + 1 ) );
						} else {
							curIdent = ( it->second ).ident;
						}
					}
				}
			} else {
				if ( curIdent == "" ) {	// ͤɬפʥϻꤷƤʤΤǡȽ
					args.push_back( argv[i] );
				} else {	// ͤɬפʥץФͤ
					OptionValue& o = option[curIdent];
					o.set( argv[i] );
					curIdent = "";
				}
			}
		}

		// ͤɬפʥФͤꤵƤʤä
		if ( curIdent != "" ) {
			fprintf( stderr, "The value for option key %s is not set. It's ignored.\n", curIdent.c_str() );
			ans = false;
		}

		return( ans );
	}

	/*
		Args::getOpt ץͤ

		const string& ident : ̻
		string& value : ͤѿ
	*/
	bool Args::getOpt( const string& ident, string& value ) const
	{
		map<string,OptionValue>::const_iterator it = option.find( ident );
		if ( it == option.end() ) {
			value = "";
			return( false );
		} else {
			value = ( it->second ).value;
			return( true );
		}
	}

	/*
		Args::getOpt ץ(true/false)(ͤʤץ)

		const string& ident : ̻
		bool& value : ͤѿ
	*/
	bool Args::getOpt( const string& ident, bool& value ) const
	{
		MapOpt::const_iterator it = option.find( ident );
		if ( it == option.end() ) {
			value = false;
			return( false );
		} else {
			value = ( it->second ).onOff;
			return( true );
		}
	}

	/*
		Args::isSet ͤåȤƤ뤫

		const string& ident : ̻
	*/
	bool Args::isSet( const string& ident ) const
	{
		MapOpt::const_iterator it = option.find( ident );
		if ( it == option.end() ) return( false );

		return( ( it->second ).onOff );
	}

	/*
		Args::keys ץ󥭡ΥꥹȤ
	*/
	vector<string> Args::keys() const
	{
		vector<string> keyList;

		map<string,string>::const_iterator it = optKey.begin();
		while ( it != optKey.end() )
			keyList.push_back( ( it++ )->first );

		return( keyList );
	}

	/*
		Args::indents ̻ҤΥꥹȤ
	*/
	vector<string> Args::idents() const
	{
		vector<string> identList;

		map<string,OptionValue>::const_iterator it = option.begin();
		while ( it != option.end() )
			identList.push_back( ( ( it++ )->second ).ident );

		return( identList );
	}

	/*
		ǥХåѽ

		string debugIdent : ǥХå٥ꤹ륪ץμ̻
		int level : ǥХå٥
		char* fmt : 
	*/
	void Args::debugPrint( string debugIdent, int level, char* fmt, ... )
	{
		va_list argp;
		string s;
		int debugLevel;
		if ( ! getOpt( debugIdent, s ) ) return;

		debugLevel = atol( s.c_str() );

		if ( level > debugLevel ) return;
		va_start( argp, fmt );
		vprintf( fmt, argp );
		va_end( argp );
	}

}
