#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <iostream>
#include <map>

#include "option.h"

using std::cout;
using std::endl;
using std::map;
using std::swap;
using std::istream;
using std::ifstream;
using namespace OptionLibrary;

// ץΥꥹ
static DefaultOptionValue optionList[] =
{
	{ "encode",	"e", "", "Ź沽Ԥ(ǥե)",			true,  false },
	{ "decode",	"d", "", "沽Ԥ",						false, false },
	{ "prime",	"p", "", "ǿΥڥꤹ",				false, true },
	{ "power",	"k", "", "ؿꤹ",					false, true },
	{ "factor",	"i", "", "ǰʬԤ",					false, true },
	{ "gcd",	"g", "", "Ȱ켡β׻",	false, true },
	{ "test",	"t", "", "RSAΥƥȼ¹",				false, false },
	{ "help",	"h", "", "إ",							false, false },
};

/*
	help : إפɽ
*/
void help()
{
	fprintf( stderr, "rsa [option] filename\n\n" );
	for ( unsigned int i = 0 ; i < sizeof( optionList ) / sizeof( optionList[0] ) ; i++ )
		fprintf( stderr, "-%s : %s\n", optionList[i].key.c_str(), optionList[i].description.c_str() );
	fprintf( stderr, "\nfilenameꤵƤʤ硢ɸϤɤ߹ߤޤ\n" );

	exit( -1 );
}

/*
	power : ֤ˡȤäˡnΤ٤׻( aknǳä; )

	unsigned int a : 
	unsigned int k : ؿ
	unsigned int n : ˡ
*/
unsigned int power( unsigned int a, unsigned int k, unsigned int n )
{
	/* n16Bitۤȱ黻̤32Bitۤ뤿 */
	if ( n >= 0x10000 ) return( 0 );

	if ( a == 0 || n == 0 ) return( 0 );	/* ˡ0ξ0֤ */
	if ( k == 0 ) return( 1 % n );			/* ؿ0ξˡnˤ1 */

	unsigned int currentMod = a % n;
	unsigned int currentValue = ( ( k & 1 ) > 0 ) ? currentMod : 1;

	for ( k >>= 1 ; k > 0 ; k >>= 1 ) {
		currentMod = ( currentMod * currentMod ) % n;
		if ( ( k & 1 ) > 0 )
			currentValue = ( currentValue * currentMod ) % n;
	}

	return( currentValue );
}

/*
	calcGCD : 桼åɤθ߽ˡȤäη׻
			  ax + by = gcd(a,b)Ʊ˵

	unsigned int a, b : GCDĤμ
	int& x, y : Ǽե

	 : 
*/
unsigned int calcGCD( unsigned int a, unsigned int b )
{
	if ( a < b ) swap( a, b );

	do {
		unsigned int r = a % b;
		a = b;
		b = r;
	} while ( b != 0 );

	return( a );
}
unsigned int calcGCD( unsigned int a, unsigned int b, int& x, int& y )
{
	x = 1;
	y = 0;
	int x1 = 0;
	int y1 = 1;

	if ( a < b ) {
		swap( a, b );
		swap( x, y );
		swap( x1, y1 );
	}

	do {
		unsigned int q = a / b;
		unsigned int r = a % b;

		int x2 = x - q * x1;
		int y2 = y - q * y1;
		x = x1;
		y = y1;
		x1 = x2;
		y1 = y2;

		a = b;
		b = r;
	} while ( b != 0 );

	return( a );
}

/*
	root : ˡnǤk躬׻( k褷nǳä;꤬bˤʤ褦ʿ )

	unsigned int k : ؿ
	unsigned int b : ;
	unsigned int n : ˡ
	unsigned int lcd : ǿ(p,q)κǾܿ
*/
unsigned int root( unsigned int k, unsigned int b, unsigned int n, unsigned int lcd )
{
	int u, v;
	if ( calcGCD( k, lcd, u, v ) != 1 ) return( 0 );

	while ( u < 0 ) u += lcd;

	return( power( b, u, n ) );
}

/*
	isPrime : ǿǤ뤫Ƚ̤

	T prime : Ƚоݤμ
	T& div : Ǿǰ
*/
template<class T> bool isPrime( T prime, T& div )
{
	if ( prime == 0 || prime == 1 ) {
		div = prime;
		return( false );
	}
	if ( prime == 2 ) {
		div = 1;
		return( true );
	}
	if ( ( prime & 1 ) == 0 ) {
		div = 2;
		return( false );
	}

	T maxNum = (T)( sqrt( (double)prime ) ) + 1;
	for ( div = 3 ; div < maxNum ; div += 2 )
		if ( prime % div == 0 ) return( false );

	div = 1;
	return( true );
}

/*
	intFactor : ǰʬ

	unsigned int num : оݤμ
	map<unsigned int,unsigned int>& mapDiv : ʬ򤷤ǰǼMap
*/
void intFactor( unsigned int num, map<unsigned int,unsigned int>& mapDiv )
{
	unsigned int div;

	mapDiv.clear();
	while ( ! isPrime( num, div ) ) {
		if ( num < 2 ) break;
		if ( mapDiv.find( div ) != mapDiv.end() )
			mapDiv[div]++;
		else
			mapDiv[div] = 1;
		num /= div;
	}

	if ( mapDiv.find( num ) != mapDiv.end() )
		mapDiv[num]++;
	else
		mapDiv[num] = 1;
}

/*
	printFactor : ǰʬ򤷤̤ɽ

	unsigned int num : оݤμ
	map<unsigned int,unsigned int>& mapDiv : ʬ򤷤ǰǼMap
*/
void printFactor( unsigned int num, const map<unsigned int,unsigned int>& mapDiv )
{
	map<unsigned int,unsigned int>::const_iterator it = mapDiv.begin();
	printf( "%d = ", num );
	while ( it != mapDiv.end() ) {
		if ( it != mapDiv.begin() ) printf( " x " );
		printf( "%d", it->first );
		if ( it->second > 1 ) printf( "^%d", it->second );
		++it;
	}
	printf( "\n" );
}

/*
	getPrimeNumber : 0255ޤǤδ֤ǥǿ֤
*/
unsigned char getPrimeNumber()
{
	unsigned char r, c;
	do {
		r = (unsigned char)( (double)rand() / ( (double)RAND_MAX + 1 ) * UCHAR_MAX ) | 1;
	} while ( ! isPrime( r, c ) );

	return( r );
}

/*
	checkPrimes : ǿΥå

	unsigned char p, q : åоݤǿ
*/
bool checkPrimes( unsigned char p, unsigned char q )
{
	unsigned char c;

	if ( ! isPrime( p, c ) ) {
		fprintf( stderr, "p=%dǿǤϤޤ\n", p );
		return( false );
	}
	if ( ! isPrime( q, c ) ) {
		fprintf( stderr, "q=%dǿǤϤޤ\n", q );
		return( false );
	}

	if ( p == q ) {
		fprintf( stderr, "pqϰۤʤǤʤФʤޤ\n" );
		return( false );
	}

	if ( p * q <= UCHAR_MAX ) {
		fprintf( stderr, "pqѤ%dʾǤʤФʤޤ\n", UCHAR_MAX + 1 );
		return( false );
	}

	return( true );
}

/*
	splitToInt : ʸĤʬ䤹

	string s : оݤʸ
	int& i, j : ʬοͤϿѿ
*/
void splitToInt( string s, int& i, int& j )
{
	string::size_type st = s.find( ',' );
	toInt( s.substr( 0, st ), i );
	toInt( s.substr( st + 1 ), j );
}

/*
	getPrimeFromOpt : 줿ǿ

	const Args& args : Υꥹ
	string key : ǿץΥ
	unsigned char p, q : ǿѿ
*/
bool getPrimeFromOpt( const Args& args, string key, unsigned char& p, unsigned char& q )
{
	int i = 0, j = 0;
	string s;

	if ( ! args.getOpt( key, s ) ) return( false );
	splitToInt( s, i, j );
	if ( i <= 0 || i > UCHAR_MAX ) {
		fprintf( stderr, "p=%dͭϰ(1-%d)ۤƤޤ\n", i, UCHAR_MAX );
		return( false );
	}
	if ( j <= 0 || j > UCHAR_MAX ) {
		fprintf( stderr, "q=%dͭϰ(1-%d)ۤƤޤ\n", j, UCHAR_MAX );
		return( false );
	}
	p = i;
	q = j;

	return( checkPrimes( p, q ) );
}

/*
	calcGCDMain : Ȱ켡β

	const Args& args : Υꥹ
*/
void calcGCDMain( const Args& args )
{
	string s;
	int a, b;
	int x, y;
	int xSign = 1, ySign = 1;

	if ( ! args.getOpt( "gcd", s ) ) exit( -1 );
	splitToInt( s, a, b );
	if ( a == 0 || b == 0 ) {
		fprintf( stderr, "켡η0ꤹ뤳ȤϤǤޤ\n" );
		exit( -1 );
	}
	if ( a < 0 ) {
		a *= -1;
		xSign = -1;
	}
	if ( b < 0 ) {
		b *= -1;
		ySign = -1;
	}

	unsigned int gcd = calcGCD( a, b, x, y );
	printf( "gcd(%d,%d) = %d\n", a, b, gcd );
	a *= xSign;
	x *= xSign;
	b *= ySign;
	y *= ySign;

	printf( "(%d)x + (%d)y = %d β x=%d, y=%d\n", a, b, gcd, x, y );

	exit( 0 );
}

/*
	intFactorMain : ǰʬ롼Υᥤ

	const Args& args : Υꥹ
*/
void intFactorMain( const Args& args )
{
	string s;
	int i;

	if ( ! args.getOpt( "factor", s ) ) exit( -1 );

	if ( ! toInt( s, i ) ) exit( -1 );
	if ( i < 0 ) exit( -1 );

	map<unsigned int,unsigned int> mapDiv;
	intFactor( i, mapDiv );
	printFactor( i, mapDiv );

	exit( 0 );
}

/*
	RSAEncode : RSAŹ沽

	istream& fin : ϥȥ꡼
	unsigned char p, q : ǿ
	unsigned int k : ؿ
	unsigned int lcd : pqκǾܿ
*/
void RSAEncode( istream& fin, unsigned char p, unsigned char q, unsigned int k, unsigned int lcd )
{
	char sc;
	unsigned int n = p * q;

	while ( fin.get( sc ) ) {
		unsigned char c = *( (unsigned char*)&sc );
		unsigned int enc = power( c, k, n );
		cout.put( (unsigned char)( ( enc >> 8 ) & 0xFF ) );
		cout.put( (unsigned char)( enc & 0xFF ) );
	}
}

/*
	RSADecode : RSA沽

	istream& fin : ϥȥ꡼
	unsigned char p, q : ǿ
	unsigned int k : ؿ
	unsigned int lcd : pqκǾܿ
*/
void RSADecode( istream& fin, unsigned char p, unsigned char q, unsigned int k, unsigned int lcd )
{
	char sc;
	unsigned int n = p * q;
	unsigned int dec = 0;
	int u, v;

	calcGCD( k, lcd, u, v );
	while ( u < 0 )
		u += lcd;

	bool evenFlg = false;
	while ( fin.get( sc ) ) {
		unsigned char c = *( (unsigned char*)&sc );
		if ( evenFlg ) {
			dec |= c;
			unsigned int i = power( dec, u, n );
			cout.put( (unsigned char)( i & 0xFF ) );
		} else {
			dec = ( c & 0xFF ) << 8;
		}
		evenFlg = ! evenFlg;
	}
}

/*
	RSATest : RSAŹ沽沽Υƥ

	unsigned char p, q : ǿ
	unsigned int k : ؿ
	unsigned int lcd : pqκǾܿ
*/
void RSATest( unsigned char p, unsigned char q, unsigned int k, unsigned int lcd )
{
	int u, v; 
	unsigned int n = p * q;

	calcGCD( k, lcd, u, v );
	while ( u < 0 ) u += lcd; 

	for ( unsigned int i = 0 ; i < n ; i++ ) {
		unsigned int enc = power( i, k, n );
		unsigned int dec = power( enc, u, n );
		cout << i << "->" << enc << "->" << dec << "\t";
		cout << ( ( i != dec ) ? "Fail" : "Pass" ) << endl;
	}
}

/*
	RSAMain : RSAŹ沽沽Υᥤ󡦥롼

	const Args& args : Υꥹ
*/
int RSAMain( const Args& args )
{
	// ǿμ
	unsigned char p, q;
	if ( args.isSet( "prime" ) ) {
		if ( ! getPrimeFromOpt( args, "prime", p, q ) )
			return( -1 );
	} else {
		do {
			p = getPrimeNumber();
			q = getPrimeNumber();
		} while ( p == q );
	}

	// pqκǾܿ
	unsigned int lcd = ( p - 1 ) * ( q - 1 ) / calcGCD( p - 1, q - 1 );

	// ؿkμ
	unsigned int k = 0;
	if ( args.isSet( "power" ) ) {
		string s;
		if ( ! args.getOpt( "power", s ) ) return( -1 );
		if ( ! toUInt( s, k ) ) return( -1 );
		if ( k == 0 ) return( -1 );
		if ( calcGCD( lcd, k ) > 1 )
			fprintf( stderr, "k=%dlcd(%d,%d)=%dϸߤǤǤʤᡢkƷ׻ޤ\n", k, p - 1, q - 1, lcd );
	}
	if ( k == 0 ) {
		do {
			k = getPrimeNumber();
		} while ( calcGCD( lcd, k ) > 1 );
	}

	// ƥȡ롼ƤӽФ
	if ( args.isSet( "test" ) ) {
		RSATest( p, q, k, lcd );
		return( 0 );
	}

	string fileName = args.getArg( 0 );

	void (*RSAFunc)( istream&, unsigned char, unsigned char, unsigned int, unsigned int );
	RSAFunc = ( args.isSet( "decode" ) ) ? RSADecode : RSAEncode;

	if ( fileName != "" ) {
		ifstream fin;
		fin.open( fileName.c_str() );
		if ( ! fin ) {
			fprintf( stderr, "եɤ߹ߤ˼Ԥޤ : %s", fileName.c_str() );
			return( -1 );
		}
		RSAFunc( fin, p, q, k, lcd );
	} else {
		RSAFunc( std::cin, p, q, k, lcd );
	}

	return( 0 );
}

/*
	ᥤ롼
*/
int main( int argc, char* argv[] )
{
	Args args( optionList, sizeof( optionList ) / sizeof( optionList[0] ) );
	args.set( argc, argv );

	if ( args.isSet( "help" ) ) help();
	if ( args.isSet( "gcd" ) ) calcGCDMain( args );
	if ( args.isSet( "factor" ) ) intFactorMain( args );

	return( RSAMain( args ) );
}
